@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2024,
+ * Copyright (C) 1986, 1988, 1989, 1991-2025,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -327,7 +327,7 @@ int
 strncasecmpmbs(const unsigned char *s1, const unsigned char *s2, size_t n)
 {
 	size_t i1, i2, mbclen1, mbclen2, gap;
-	wchar_t wc1, wc2;
+	char32_t wc1, wc2;
 	mbstate_t mbs1, mbs2;
 
 	memset(& mbs1, 0, sizeof(mbs1));
@@ -338,7 +338,7 @@ strncasecmpmbs(const unsigned char *s1, const unsigned char *s2, size_t n)
 			mbclen1 = 1;
 			wc1 = btowc_cache(s1[i1]);
 		} else {
-			mbclen1 = mbrtowc(& wc1, (const char *)s1 + i1,
+			mbclen1 = mbrtoc32(& wc1, (const char *)s1 + i1,
 					  n - i1, & mbs1);
 			if (mbclen1 == (size_t) -1 || mbclen1 == (size_t) -2 || mbclen1 == 0) {
 				/* We treat it as a singlebyte character. */
@@ -350,7 +350,7 @@ strncasecmpmbs(const unsigned char *s1, const unsigned char *s2, size_t n)
 			mbclen2 = 1;
 			wc2 = btowc_cache(s2[i2]);
 		} else {
-			mbclen2 = mbrtowc(& wc2, (const char *)s2 + i2,
+			mbclen2 = mbrtoc32(& wc2, (const char *)s2 + i2,
 					  n - i2, & mbs2);
 			if (mbclen2 == (size_t) -1 || mbclen2 == (size_t) -2 || mbclen2 == 0) {
 				/* We treat it as a singlebyte character. */
@@ -465,7 +465,7 @@ do_index(int nargs)
 			if (l2 > l1)
 				break;
 			if (! do_single_byte && gawk_mb_cur_max > 1) {
-				const wchar_t *pos;
+				const char32_t *pos;
 
 				pos = wcasestrstr(s1->wstptr, s1->wstlen, s2->wstptr, s2->wstlen);
 				if (pos == NULL)
@@ -497,7 +497,7 @@ do_index(int nargs)
 				break;
 			}
 			if (! do_single_byte && gawk_mb_cur_max > 1) {
-				const wchar_t *pos;
+				const char32_t *pos;
 
 				pos = wstrstr(s1->wstptr, s1->wstlen, s2->wstptr, s2->wstlen);
 				if (pos == NULL)
@@ -559,6 +559,14 @@ do_isarray(int nargs)
 	check_exact_args(nargs, "isarray", 1);
 
 	tmp = POP();
+
+	if (tmp->type == Node_param_list) {
+		tmp = GET_PARAM(tmp->param_cnt);
+		if (tmp->type == Node_array_ref) {
+			tmp = tmp->orig_array;
+		}
+	}
+
 	if (tmp->type != Node_var_array) {
 		ret = 0;
 		// could be Node_var_new
@@ -579,16 +587,10 @@ do_length(int nargs)
 	check_exact_args(nargs, "length", 1);
 
 	tmp = POP();
-	if (tmp->type == Node_var_array) {
-		static bool warned = false;
-		unsigned long size;
 
-		if (do_posix)
-			fatal(_("length: received array argument"));
-   		if (do_lint_extensions && ! warned) {
-			warned = true;
-			lintwarn(_("`length(array)' is a gawk extension"));
-		}
+	// As of the 2024 POSIX standard, length(array) is standardized. Yay!
+	if (tmp->type == Node_var_array) {
+		unsigned long size;
 
 		/*
 		 * Support for deferred loading of array elements requires that
@@ -724,12 +726,12 @@ do_substr(int nargs)
 			_("substr: non-integer length %g will be truncated"),
 					d_length);
 
-			if (d_length > SIZE_MAX)
+			if (d_length > (double) SIZE_MAX)
 				lintwarn(
 			_("substr: length %g too big for string indexing, truncating to %g"),
 					d_length, (double) SIZE_MAX);
 		}
-		if (d_length < SIZE_MAX)
+		if (d_length < (double) SIZE_MAX)
 			length = d_length;
 		else
 			length = SIZE_MAX;
@@ -747,7 +749,7 @@ do_substr(int nargs)
 			 d_index);
 
 	/* awk indices are from 1, C's are from 0 */
-	if (d_index <= SIZE_MAX)
+	if (d_index <= (double) SIZE_MAX)
 		indx = d_index - 1;
 	else
 		indx = SIZE_MAX;
@@ -800,7 +802,7 @@ do_substr(int nargs)
 	else {
 		/* multibyte case, more work */
 		size_t result;
-		wchar_t *wp;
+		char32_t *wp;
 		mbstate_t mbs;
 		char *substr, *cp;
 
@@ -810,10 +812,10 @@ do_substr(int nargs)
 		 * way to do things.
 		 */
 		memset(& mbs, 0, sizeof(mbs));
-		emalloc(substr, char *, (length * gawk_mb_cur_max) + 1, "do_substr");
+		emalloc(substr, char *, (length * gawk_mb_cur_max) + 1);
 		wp = t1->wstptr + indx;
 		for (cp = substr; length > 0; length--) {
-			result = wcrtomb(cp, *wp, & mbs);
+			result = c32rtomb(cp, *wp, & mbs);
 			if (result == (size_t) -1)	/* what to do? break seems best */
 				break;
 			cp += result;
@@ -951,9 +953,9 @@ do_strftime(int nargs)
 			break;
 		bufsize *= 2;
 		if (bufp == buf)
-			emalloc(bufp, char *, bufsize, "do_strftime");
+			emalloc(bufp, char *, bufsize);
 		else
-			erealloc(bufp, char *, bufsize, "do_strftime");
+			erealloc(bufp, char *, bufsize);
 	}
 	ret = make_string(bufp, buflen);
 	if (bufp != buf)
@@ -1095,20 +1097,14 @@ do_system(int nargs)
 		 * divides the result by 256.  That normally gives the
 		 * exit status but gives a weird result for death-by-signal.
 		 * So we compromise as follows:
+		 *
+		 * 7/2025. BWK awk now does what we do to sanitize the
+		 * status.  I removed the MinGW code for do_traditional
+		 * since we no longer care about that option here.
 		 */
 		ret = status;
-		if (status != -1) {
-			if (do_posix)
-				;	/* leave it alone, full 16 bits */
-			else if (do_traditional)
-#ifdef __MINGW32__
-				ret = (((unsigned)status) & ~0xC0000000);
-#else
-				ret = (status / 256.0);
-#endif
-			else
-				ret = sanitize_exit_status(status);
-		}
+		if (status != -1 && ! do_posix)
+			ret = sanitize_exit_status(status);
 
 		if ((BINMODE & BINMODE_INPUT) != 0)
 			os_setbinmode(fileno(stdin), O_BINARY);
@@ -1258,7 +1254,7 @@ do_print_rec(int nargs, int redirtype)
 /* is_wupper --- function version of iswupper for passing function pointers */
 
 static int
-is_wupper(wchar_t c)
+is_wupper(char32_t c)
 {
 	return iswupper(c);
 }
@@ -1266,7 +1262,7 @@ is_wupper(wchar_t c)
 /* is_wlower --- function version of iswlower for passing function pointers */
 
 static int
-is_wlower(wchar_t c)
+is_wlower(char32_t c)
 {
 	return iswlower(c);
 }
@@ -1274,7 +1270,7 @@ is_wlower(wchar_t c)
 /* to_wupper --- function version of towupper for passing function pointers */
 
 static int
-to_wlower(wchar_t c)
+to_wlower(char32_t c)
 {
 	return towlower(c);
 }
@@ -1282,7 +1278,7 @@ to_wlower(wchar_t c)
 /* to_wlower --- function version of towlower for passing function pointers */
 
 static int
-to_wupper(wchar_t c)
+to_wupper(char32_t c)
 {
 	return towupper(c);
 }
@@ -1290,13 +1286,13 @@ to_wupper(wchar_t c)
 /* wide_change_case --- generic case converter for wide characters */
 
 static void
-wide_change_case(wchar_t *wstr,
+wide_change_case(char32_t *wstr,
 			size_t wlen,
-			int (*is_x)(wchar_t c),
-			int (*to_y)(wchar_t c))
+			int (*is_x)(char32_t c),
+			int (*to_y)(char32_t c))
 {
 	size_t i;
-	wchar_t *wcp;
+	char32_t *wcp;
 
 	for (i = 0, wcp = wstr; i < wlen; i++, wcp++)
 		if (is_x(*wcp))
@@ -1306,7 +1302,7 @@ wide_change_case(wchar_t *wstr,
 /* wide_toupper --- map a wide string to upper case */
 
 static void
-wide_toupper(wchar_t *wstr, size_t wlen)
+wide_toupper(char32_t *wstr, size_t wlen)
 {
 	wide_change_case(wstr, wlen, is_wlower, to_wupper);
 }
@@ -1314,7 +1310,7 @@ wide_toupper(wchar_t *wstr, size_t wlen)
 /* wide_tolower --- map a wide string to lower case */
 
 static void
-wide_tolower(wchar_t *wstr, size_t wlen)
+wide_tolower(char32_t *wstr, size_t wlen)
 {
 	wide_change_case(wstr, wlen, is_wupper, to_wlower);
 }
@@ -1554,12 +1550,12 @@ do_srand(int nargs)
 	check_args_min_max(nargs, "srand", 0, 1);
 
 	if (nargs == 0)
-		srandom((unsigned int) (save_seed = (long) time((time_t *) 0)));
+		srandom((unsigned int) (save_seed = (unsigned long) time((time_t *) 0)));
 	else {
 		tmp = POP_SCALAR();
 		if (do_lint && (fixtype(tmp)->flags & NUMBER) == 0)
 			lintwarn(_("%s: received non-numeric argument"), "srand");
-		srandom((unsigned int) (save_seed = (long) force_number(tmp)->numbr));
+		srandom((unsigned int) (save_seed = (unsigned long) force_number(tmp)->numbr));
 		DEREF(tmp);
 	}
 	return make_number((AWKNUM) ret);
@@ -1577,7 +1573,7 @@ do_match(int nargs)
 	int rstart, len, ii;
 	int rlength;
 	Regexp *rp;
-	regoff_t s;
+	ssize_t s;
 	char *start;
 	char *buf = NULL;
 	char buff[100];
@@ -1602,7 +1598,7 @@ do_match(int nargs)
 	if (do_lint && (fixtype(t1)->flags & (STRING|USER_INPUT)) == 0)
 		lintwarn(_("%s: received non-string first argument"), "match");
 
-	rstart = research(rp, t1->stptr, 0, t1->stlen, RE_NEED_START);
+	rstart = research(rp, t1->stptr, 0, t1->stlen, RE_NEED_START|RE_NEED_SUB);
 	if (rstart >= 0) {	/* match succeded */
 		size_t *wc_indices = NULL;
 
@@ -1646,9 +1642,9 @@ do_match(int nargs)
 					amt = ilen + subseplen + strlen("length") + 1;
 
 					if (oldamt == 0) {
-						emalloc(buf, char *, amt, "do_match");
+						emalloc(buf, char *, amt);
 					} else if (amt > oldamt) {
-						erealloc(buf, char *, amt, "do_match");
+						erealloc(buf, char *, amt);
 					}
 					oldamt = amt;
 					memcpy(buf, buff, ilen);
@@ -1679,6 +1675,14 @@ do_match(int nargs)
 	}
 
 	DEREF(t1);
+	// Someone used $0 as the second parameter. We have to treat
+	// it with extra TLC.
+	if (tre->type == Node_dynregex
+	    && tre->re_exp->valref > 1
+	    && (tre->re_exp->flags & MALLOC) == 0) {
+		unref(tre->re_exp);
+		tre->re_exp = NULL;
+	}
 	unref(RSTART_node->var_value);
 	RSTART_node->var_value = make_number((AWKNUM) rstart);
 	unref(RLENGTH_node->var_value);
@@ -1814,6 +1818,7 @@ do_sub(int nargs, unsigned int flags)
 	long current;
 	bool lastmatchnonzero;
 	char *mb_indices = NULL;
+	int searchflags = RE_NEED_START;
 
 	if ((flags & GENSUB) != 0) {
 		double d;
@@ -1836,7 +1841,7 @@ do_sub(int nargs, unsigned int flags)
 			d = get_number_d(glob_flag);
 			if (d < 1)
 				how_many = 1;
-			else if (d < LONG_MAX)
+			else if (d < (double) LONG_MAX)
 				how_many = d;
 			else
 				how_many = LONG_MAX;
@@ -1848,6 +1853,7 @@ do_sub(int nargs, unsigned int flags)
 			}
 		}
 		DEREF(glob_flag);
+		searchflags |= RE_NEED_SUB;
 	} else {
 		if ((flags & GSUB) != 0) {
 			check_exact_args(nargs, "gsub", 3);
@@ -1879,7 +1885,7 @@ do_sub(int nargs, unsigned int flags)
 	decr_sp();		/* regexp, already updated above */
 
 	/* do the search early to avoid work on non-match */
-	if (research(rp, target->stptr, 0, target->stlen, RE_NEED_START) == -1 ||
+	if (research(rp, target->stptr, 0, target->stlen, searchflags) == -1 ||
 			RESTART(rp, target->stptr) > target->stlen)
 		goto done;
 
@@ -1901,7 +1907,7 @@ do_sub(int nargs, unsigned int flags)
 	 * for example.
 	 */
 	if (gawk_mb_cur_max > 1 && repllen > 0) {
-		emalloc(mb_indices, char *, repllen * sizeof(char), "do_sub");
+		emalloc(mb_indices, char *, repllen * sizeof(char));
 		index_multibyte_buffer(repl, mb_indices, repllen);
 	}
 
@@ -1954,7 +1960,7 @@ do_sub(int nargs, unsigned int flags)
 
 	/* guesstimate how much room to allocate; +1 forces > 0 */
 	buflen = textlen + (ampersands + 1) * repllen + 1;
-	emalloc(buf, char *, buflen + 1, "do_sub");
+	emalloc(buf, char *, buflen + 1);
 	buf[buflen] = '\0';
 
 	bp = buf;
@@ -1981,7 +1987,7 @@ do_sub(int nargs, unsigned int flags)
 		sofar = bp - buf;
 		while (buflen < (sofar + len + 1)) {
 			buflen *= 2;
-			erealloc(buf, char *, buflen, "sub_common");
+			erealloc(buf, char *, buflen);
 			bp = buf + sofar;
 		}
 		for (scan = text; scan < matchstart; scan++)
@@ -2107,14 +2113,16 @@ do_sub(int nargs, unsigned int flags)
 
 		if ((current >= how_many && ! global)
 		    || ((long) textlen <= 0 && matchstart == matchend)
-		    || research(rp, target->stptr, text - target->stptr, textlen, RE_NEED_START) == -1)
+		    || research(rp, target->stptr, text - target->stptr,
+			    use_gnu_matchers ? textlen : target->stlen,
+			    RE_NEED_START) == -1)
 			break;
 
 	}
 	sofar = bp - buf;
 	if (buflen < (sofar + textlen + 1)) {
 		buflen = sofar + textlen + 1;
-		erealloc(buf, char *, buflen, "do_sub");
+		erealloc(buf, char *, buflen);
 		bp = buf + sofar;
 	}
 	/*
@@ -2170,20 +2178,8 @@ done:
 		 * remain a regexp. In that case, we have to update the compiled
 		 * regular expression that it holds.
 		 */
-		bool is_regex = false;
-		NODE *target = *lhs;
+		bool is_regex = ((target->flags & REGEX) != 0);
 
-		if ((target->flags & REGEX) != 0) {
-			is_regex = true;
-
-			if (target->valref == 1) {
-				// free old regex registers
-				refree(target->typed_re->re_reg[0]);
-				if (target->typed_re->re_reg[1] != NULL)
-					refree(target->typed_re->re_reg[1]);
-				freenode(target->typed_re);
-			}
-		}
 		unref(*lhs);		// nuke original value
 		if (is_regex)
 			*lhs = make_typed_regex(buf, textlen);
@@ -2204,9 +2200,13 @@ call_sub(const char *name, int nargs)
 	NODE **lhs, *rhs;
 	NODE *zero = make_number(0.0);
 	NODE *result;
+	const char *fname = name;
 
-	if (name[0] == 'g') {
-		if (name[1] == 'e')
+	if (fname[0] == 'a')	// awk::...
+		fname += 5;
+
+	if (fname[0] == 'g') {
+		if (fname[1] == 'e')
 			flags = GENSUB;
 		else
 			flags = GSUB;
@@ -2307,17 +2307,33 @@ call_match(int nargs)
 	if (nargs == 3)
 		array = POP();
 	regex = POP();
-
-	/* Don't need to pop the string just to push it back ... */
+	text = POP();
 
 	bool need_free = false;
 	if ((regex->flags & REGEX) != 0)
 		regex = regex->typed_re;
-	else {
+	else if (regex->type == Node_var_new || regex->type == Node_elem_new) {
+		if (regex->type == Node_elem_new)
+			elem_new_reset(regex);
+		else if (regex->vname != NULL)
+			efree(regex->vname);
+		memset(regex, 0, sizeof(*regex));
+		regex->type = Node_dynregex;
+		regex->re_exp = dupnode(Nnull_string);
+	} else {
 		regex = make_regnode(Node_regex, regex);
 		need_free = true;
 	}
 
+	if (text->type == Node_var_new || text->type == Node_elem_new) {
+		if (text->type == Node_elem_new)
+			elem_new_reset(text);
+		else if (text->vname != NULL)
+			efree(text->vname);
+		text = dupnode(Nnull_string);
+	}
+
+	PUSH(text);
 	PUSH(regex);
 
 	if (array)
@@ -2342,12 +2358,16 @@ call_split_func(const char *name, int nargs)
 {
 	NODE *regex, *seps;
 	NODE *result;
+	const char *fname = name;
 
 	regex = seps = NULL;
 	if (nargs < 2 || nargs > 4)
 		fatal(_("indirect call to %s requires two to four arguments"),
 				name);
 
+	if (fname[0] == 'a')	// awk::...
+		fname += 5;
+
 	if (nargs == 4)
 		seps = POP();
 
@@ -2361,7 +2381,7 @@ call_split_func(const char *name, int nargs)
 			need_free = true;
 		}
 	} else {
-		if (name[0] == 's') {
+		if (fname[0] == 's') {
 			regex = make_regnode(Node_regex, FS_node->var_value);
 			regex->re_flags |= FS_DFLT;
 		} else
@@ -2378,7 +2398,7 @@ call_split_func(const char *name, int nargs)
 	if (seps)
 		PUSH(seps);
 
-	result = (name[0] == 's') ? do_split(nargs) : do_patsplit(nargs);
+	result = (fname[0] == 's') ? do_split(nargs) : do_patsplit(nargs);
 
 	if (need_free) {
 		refree(regex->re_reg[0]);
@@ -2424,20 +2444,25 @@ do_lshift(int nargs)
 	if (val < 0 || shift < 0)
 		fatal(_("lshift(%f, %f): negative values are not allowed"), val, shift);
 
-	if (do_lint) {
-		if (double_to_int(val) != val || double_to_int(shift) != shift)
-			lintwarn(_("lshift(%f, %f): fractional values will be truncated"), val, shift);
-		if (shift >= sizeof(uintmax_t) * CHAR_BIT)
-			lintwarn(_("lshift(%f, %f): too large shift value will give strange results"), val, shift);
+	if (do_lint && (double_to_int(val) != val || double_to_int(shift) != shift))
+		lintwarn(_("lshift(%f, %f): fractional values will be truncated"), val, shift);
+
+	if (shift < sizeof(uintmax_t) * CHAR_BIT) {
+		// within range
+		uval = (uintmax_t) val;
+		ushift = (uintmax_t) shift;
+
+		res = uval << ushift;
+	} else {
+		// out of range
+		if (do_lint)
+			lintwarn(_("lshift(%f, %f): too large shift value returns zero"), val, shift);
+		res = 0;
 	}
 
 	DEREF(s1);
 	DEREF(s2);
 
-	uval = (uintmax_t) val;
-	ushift = (uintmax_t) shift;
-
-	res = uval << ushift;
 	return make_integer(res);
 }
 
@@ -2465,20 +2490,25 @@ do_rshift(int nargs)
 	if (val < 0 || shift < 0)
 		fatal(_("rshift(%f, %f): negative values are not allowed"), val, shift);
 
-	if (do_lint) {
-		if (double_to_int(val) != val || double_to_int(shift) != shift)
-			lintwarn(_("rshift(%f, %f): fractional values will be truncated"), val, shift);
-		if (shift >= sizeof(uintmax_t) * CHAR_BIT)
-			lintwarn(_("rshift(%f, %f): too large shift value will give strange results"), val, shift);
+	if (do_lint && (double_to_int(val) != val || double_to_int(shift) != shift))
+		lintwarn(_("rshift(%f, %f): fractional values will be truncated"), val, shift);
+
+	if (shift < sizeof(uintmax_t) * CHAR_BIT) {
+		// within range
+		uval = (uintmax_t) val;
+		ushift = (uintmax_t) shift;
+
+		res = uval >> ushift;
+	} else {
+		// out of range
+		if (do_lint)
+			lintwarn(_("rshift(%f, %f): too large shift value returns zero"), val, shift);
+		res = 0;
 	}
 
 	DEREF(s1);
 	DEREF(s2);
 
-	uval = (uintmax_t) val;
-	ushift = (uintmax_t) shift;
-
-	res = uval >> ushift;
 	return make_integer(res);
 }
 
@@ -2640,6 +2670,8 @@ nondec2awknum(char *str, size_t len, char **endptr)
 	char save;
 	short val;
 	char *start = str;
+	size_t savelen = len;
+	bool is_hex_float = false;
 
 	if (len >= 2 && *str == '0' && (str[1] == 'x' || str[1] == 'X')) {
 		/*
@@ -2682,6 +2714,14 @@ nondec2awknum(char *str, size_t len, char **endptr)
 			case 'F':
 				val = *str - 'A' + 10;
 				break;
+			case 'p':
+			case 'P':
+			case '.':
+				// restore the string
+				str = start;
+				len = savelen;
+				is_hex_float = true;
+				goto floating_point_hex;
 			default:
 				if (endptr)
 					*endptr = str;
@@ -2711,8 +2751,13 @@ nondec2awknum(char *str, size_t len, char **endptr)
 			*endptr = str;
 	} else {
 decimal:
+floating_point_hex:
 		save = str[len];
 		str[len] = '\0';
+		if (is_hex_float
+		    && strchr(str, 'p') == NULL 
+		    && strchr(str, 'P') == NULL)
+			fatal(_("hexadecimal floating point constants require an exponent"));
 		retval = strtod(str, endptr);
 		str[len] = save;
 	}
@@ -3032,9 +3077,9 @@ do_bindtextdomain(int nargs)
 /*
  * We define the semantics as:
  * 	numerator = int(numerator)
- *	denominator = int(denonmator)
- *	quotient = int(numerator / denomator)
- *	remainder = int(numerator % denomator)
+ *	denominator = int(denominator)
+ *	quotient = int(numerator / denominator)
+ *	remainder = int(numerator % denominator)
  */
 
 NODE *
@@ -3113,6 +3158,14 @@ do_typeof(int nargs)
 	else
 		dbg = NULL;
 	arg = POP();
+
+	if (arg->type == Node_param_list) {
+		arg = GET_PARAM(arg->param_cnt);
+		if (arg->type == Node_array_ref) {
+			arg = arg->orig_array;
+		}
+	}
+
 	switch (arg->type) {
 	case Node_var_array:
 		/* Node_var_array is never UPREF'ed */
@@ -3146,7 +3199,7 @@ do_typeof(int nargs)
 
 #define SETVAL(X, V) {	\
 	size_t l = nl + sizeof(#X);	\
-	emalloc(p, char *, l+1, "do_typeof");	\
+	emalloc(p, char *, l+1);	\
 	sprintf(p, "%s_" #X, nextfree[i].name);	\
 	assoc_set(dbg, make_str_node(p, l, ALREADY_MALLOCED), make_number((AWKNUM) (V)));	\
 }
@@ -3214,9 +3267,10 @@ do_typeof(int nargs)
 		}
 		break;
 	case Node_var_new:
+		deref = false;
+		// fall through
 	case Node_elem_new:
 		res = "untyped";
-		deref = false;
 		break;
 	case Node_array_ref:
 		/*
@@ -3301,6 +3355,13 @@ do_mkbool(int nargs)
 
 /* gawk_system --- specialized version for gawk */
 
+/*
+ * 7/2025: This got messier because we need to ignore SIGINT and SIGQUIT
+ * and block SIGCHLD while running (see the POSIX description of system()).
+ * Blocking is only possible with the POSIX sigaction/sigprocmask APIs.
+ * We do our best here, but it ain't pretty.
+ */
+
 int
 gawk_system(const char *command)
 {
@@ -3310,38 +3371,137 @@ gawk_system(const char *command)
 	pid_t childpid;
 	int status;
 
+	// In the parent, before the fork
+#ifdef HAVE_SIGPROCMASK
+	sigset_t set, oldset;
+
+	sigemptyset(& set);
+	sigaddset(& set, SIGCHLD);
+	sigprocmask(SIG_BLOCK, & set, & oldset);
+
+	struct sigaction action, old_int_action, old_quit_action;
+
+	sigemptyset(& action.sa_mask);
+	action.sa_flags = 0;
+	action.sa_handler = SIG_IGN;
+	sigaction(SIGINT, & action, & old_int_action);
+	sigaction(SIGQUIT, & action, & old_quit_action);
+#else
+	void (*istat)(int), (*qstat)(int);
+
+	istat = signal(SIGINT, SIG_IGN);
+	qstat = signal(SIGQUIT, SIG_IGN);
+#endif
+
 	if ((childpid = fork()) == 0) {
 		// child
 		set_sigpipe_to_default();
+		// in the child, restore defaults
+#ifdef HAVE_SIGPROCMASK
+		sigprocmask(SIG_SETMASK, & oldset, NULL);
+		sigaction(SIGINT, & old_int_action, NULL);
+		sigaction(SIGQUIT, & old_quit_action, NULL);
+#else
+		istat = signal(SIGINT, istat);
+		qstat = signal(SIGQUIT, qstat);
+#endif
 		execl("/bin/sh", "sh", "-c", command, NULL);
 		_exit(errno == ENOENT ? 127 : 126);
+
+		return 0;	// silence compiler warnings
 	} else {
 		// parent
 		status = wait_any(childpid);
 
+		// in the parent, restore stuff after getting the status
+#ifdef HAVE_SIGPROCMASK
+		sigprocmask(SIG_SETMASK, & oldset, NULL);
+		sigaction(SIGINT, & old_int_action, NULL);
+		sigaction(SIGQUIT, & old_quit_action, NULL);
+#else
+		istat = signal(SIGINT, istat);
+		qstat = signal(SIGQUIT, qstat);
+#endif
+
 		return status;
 	}
-#endif /* defined(VMS) || defined(__MINGW32__) */
+#endif /* ! (defined(VMS) || defined(__MINGW32__)) */
 }
 
-#if 0
-// test program
+/* do_dump_node --- dump the offsets of the NODE struct */
 
-int main(int argc, char **argv)
+NODE *
+do_dump_node(int nargs)
 {
-	struct lconv *l;
+	printf("MB_CUR_MAX = %u\n", (int) MB_CUR_MAX);
+	printf("sizeof(int) = %zu\n", sizeof(int));
+	printf("sizeof(long) = %zu\n", sizeof(long));
+	printf("sizeof(long long) = %zu\n", sizeof(long long));
+	printf("sizeof(void *) = %zu\n", sizeof(void *));
+	printf("\n");
+	fflush(stdout);
 
-	setlocale(LC_ALL, "");
-	l = localeconv();
+	FILE *fp = popen("sort -n", "w");
+	if (fp == NULL)
+		return Nnull_string;
 
-	const char *new = add_thousands("12345678901234567890.54321", l);
-	printf("%s\n", new);
-	free((void*) new);
-
-	new = add_thousands("12345678901234567890", l);
-	printf("%s\n", new);
-	free((void*) new);
-
-	return 0;
-}
+	fprintf(fp, "%zu\tarray_base\n", offsetof(NODE, array_base));
+	fprintf(fp, "%zu\tarray_capacity\n", offsetof(NODE, array_capacity));
+	fprintf(fp, "%zu\tarray_funcs\n", offsetof(NODE, array_funcs));
+	fprintf(fp, "%zu\tarray_size\n", offsetof(NODE, array_size));
+	fprintf(fp, "%zu\tbuckets\n", offsetof(NODE, buckets));
+	fprintf(fp, "%zu\tcode_ptr\n", offsetof(NODE, code_ptr));
+	fprintf(fp, "%zu\tcomment_type\n", offsetof(NODE, comment_type));
+	fprintf(fp, "%zu\tcur_idx	\n", offsetof(NODE, cur_idx	));
+	fprintf(fp, "%zu\tdup_ent\n", offsetof(NODE, dup_ent));
+	fprintf(fp, "%zu\telemnew_parent\n", offsetof(NODE, elemnew_parent));
+	fprintf(fp, "%zu\telemnew_vname\n", offsetof(NODE, elemnew_vname));
+	fprintf(fp, "%zu\tflags\n", offsetof(NODE, flags));
+	fprintf(fp, "%zu\tfor_array\n", offsetof(NODE, for_array));
+	fprintf(fp, "%zu\tfor_list\n", offsetof(NODE, for_list));
+	fprintf(fp, "%zu\tfor_list_size\n", offsetof(NODE, for_list_size));
+	fprintf(fp, "%zu\tfparms\n", offsetof(NODE, fparms));
+	fprintf(fp, "%zu\tfunc_node\n", offsetof(NODE, func_node));
+	fprintf(fp, "%zu\tlnode\n", offsetof(NODE, lnode));
+#ifdef HAVE_MPFR
+	fprintf(fp, "%zu\tmpg_i\n", offsetof(NODE, mpg_i));
+	fprintf(fp, "%zu\tmpg_numbr\n", offsetof(NODE, mpg_numbr));
 #endif
+	fprintf(fp, "%zu\tnodes\n", offsetof(NODE, nodes));
+	fprintf(fp, "%zu\tnumbr\n", offsetof(NODE, numbr));
+	fprintf(fp, "%zu\torig_array\n", offsetof(NODE, orig_array));
+	fprintf(fp, "%zu\tparam\n", offsetof(NODE, param));
+	fprintf(fp, "%zu\tparam_cnt\n", offsetof(NODE, param_cnt));
+	fprintf(fp, "%zu\tparent_array\n", offsetof(NODE, parent_array));
+	fprintf(fp, "%zu\tprev_array\n", offsetof(NODE, prev_array));
+	fprintf(fp, "%zu\tprev_frame_size\n", offsetof(NODE, prev_frame_size));
+	fprintf(fp, "%zu\tre_cnt\n", offsetof(NODE, re_cnt));
+	fprintf(fp, "%zu\tre_exp\n", offsetof(NODE, re_exp));
+	fprintf(fp, "%zu\tre_flags\n", offsetof(NODE, re_flags));
+	fprintf(fp, "%zu\tre_reg\n", offsetof(NODE, re_reg));
+	fprintf(fp, "%zu\tre_text\n", offsetof(NODE, re_text));
+	fprintf(fp, "%zu\treti\n", offsetof(NODE, reti));
+	fprintf(fp, "%zu\trnode\n", offsetof(NODE, rnode));
+	fprintf(fp, "%zu\tstack\n", offsetof(NODE, stack));
+	fprintf(fp, "%zu\tstfmt\n", offsetof(NODE, stfmt));
+	fprintf(fp, "%zu\tstlen\n", offsetof(NODE, stlen));
+	fprintf(fp, "%zu\tstptr\n", offsetof(NODE, stptr));
+#ifdef HAVE_MPFR
+	fprintf(fp, "%zu\tstrndmode\n", offsetof(NODE, strndmode));
+#endif
+	fprintf(fp, "%zu\ttable_size\n", offsetof(NODE, table_size));
+	fprintf(fp, "%zu\ttype\n", offsetof(NODE, type));
+	fprintf(fp, "%zu\ttyped_re\n", offsetof(NODE, typed_re));
+	fprintf(fp, "%zu\tvalref\n", offsetof(NODE, valref));
+	fprintf(fp, "%zu\tvar_assign\n", offsetof(NODE, var_assign));
+	fprintf(fp, "%zu\tvar_update\n", offsetof(NODE, var_update));
+	fprintf(fp, "%zu\tvar_value\n", offsetof(NODE, var_value));
+	fprintf(fp, "%zu\tvname\n", offsetof(NODE, vname));
+	fprintf(fp, "%zu\twstlen\n", offsetof(NODE, wstlen));
+	fprintf(fp, "%zu\twstptr\n", offsetof(NODE, wstptr));
+	fprintf(fp, "%zu\txarray\n", offsetof(NODE, xarray));
+
+	pclose(fp);
+	
+	return Nnull_string;
+}
