@@ -7,7 +7,7 @@
  */
 
 /*
- * Copyright (C) 1995 - 2001, 2003-2014, 2016-2018,
+ * Copyright (C) 1995 - 2001, 2003-2014, 2016-2020, 2022, 2025,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -31,8 +31,55 @@
 #include "awk.h"
 extern SRCFILE *srcfiles;
 
+struct extension {
+	const char *name;
+	const char *path;
+	struct extension *next;
+} *extension_list = NULL;
+static struct extension *last_ext = NULL;
+
 #ifdef DYNAMIC
 
+/* search_ext_list --- validate that extension still comes from the same place */
+
+static void
+search_ext_list(const char *name, const char *lib_name)
+{
+	struct extension *p;
+
+	for (p = extension_list; p != NULL; p = p->next) {
+		if (strcmp(p->name, name) == 0) {
+			if (strcmp(p->path, lib_name) != 0)
+				fatal(_("extension `%s': trying to load %s, was previously loaded from %s"),
+						p->name, lib_name, p->path);
+			else
+				return;
+		}
+	}
+
+	// append to list, since the front of the list is what's saved in the backing store. ugh.
+	if (p == NULL) {
+		emalloc(p, struct extension *, sizeof(struct extension));
+		p->name = estrdup(name, strlen(name));
+		p->path = estrdup(lib_name, strlen(lib_name));
+		last_ext->next = p;
+		last_ext = p;
+	}
+}
+
+/* init_extension_list --- put a dummy object on the list for PMA */
+
+void
+init_extension_list(void)
+{
+	struct extension *p;
+	emalloc(p, struct extension *, sizeof(struct extension));
+	p->name = estrdup("@dummy@", strlen("@dummy@"));
+	p->path = estrdup("/no/such/path", strlen("/no/such/path"));
+	p->next = NULL;
+	last_ext = extension_list = p;
+}
+
 #include <dlfcn.h>
 
 #define INIT_FUNC	"dl_load"
@@ -40,36 +87,40 @@ extern SRCFILE *srcfiles;
 /* load_ext --- load an external library */
 
 void
-load_ext(const char *lib_name)
+load_ext(const char *name, const char *lib_name)
 {
 	int (*install_func)(const gawk_api_t *const, awk_ext_id_t);
 	void *dl;
 	int flags = RTLD_LAZY;
 	int *gpl_compat;
+	bool override = (getenv("MAGIC_XYZZY") != NULL);
 
 	if (do_sandbox)
 		fatal(_("extensions are not allowed in sandbox mode"));
 
-	if (do_traditional || do_posix)
+	if (using_persistent_malloc)
+		search_ext_list(name, lib_name);	// fatal out if problem
+
+	if (do_traditional && ! override)
 		fatal(_("-l / @load are gawk extensions"));
 
 	if (lib_name == NULL)
 		fatal(_("load_ext: received NULL lib_name"));
 
 	if ((dl = dlopen(lib_name, flags)) == NULL)
-		fatal(_("load_ext: cannot open library `%s' (%s)"), lib_name,
+		fatal(_("load_ext: cannot open library `%s': %s"), lib_name,
 		      dlerror());
 
 	/* Per the GNU Coding standards */
 	gpl_compat = (int *) dlsym(dl, "plugin_is_GPL_compatible");
 	if (gpl_compat == NULL)
-		fatal(_("load_ext: library `%s': does not define `plugin_is_GPL_compatible' (%s)"),
+		fatal(_("load_ext: library `%s': does not define `plugin_is_GPL_compatible': %s"),
 				lib_name, dlerror());
 
 	install_func = (int (*)(const gawk_api_t *const, awk_ext_id_t))
 				dlsym(dl, INIT_FUNC);
 	if (install_func == NULL)
-		fatal(_("load_ext: library `%s': cannot call function `%s' (%s)"),
+		fatal(_("load_ext: library `%s': cannot call function `%s': %s"),
 				lib_name, INIT_FUNC, dlerror());
 
 	if (install_func(& api_impl, NULL /* ext_id */) == 0)
@@ -97,7 +148,7 @@ make_builtin(const char *name_space, const awk_ext_func_t *funcinfo)
 	assert(name_space != NULL);
 	if (name_space[0] == '\0' || strcmp(name_space, awk_namespace) == 0) {
 		if (check_special(name) >= 0)
-			fatal(_("make_builtin: can't use gawk built-in `%s' as function name"), name);
+			fatal(_("make_builtin: cannot use gawk built-in `%s' as function name"), name);
 
 		f = lookup(name);
 		install_name = estrdup(name, strlen(name));
@@ -106,13 +157,13 @@ make_builtin(const char *name_space, const awk_ext_func_t *funcinfo)
 			return awk_false;
 
 		if (check_special(name_space) >= 0)
-			fatal(_("make_builtin: can't use gawk built-in `%s' as namespace name"), name_space);
+			fatal(_("make_builtin: cannot use gawk built-in `%s' as namespace name"), name_space);
 		if (check_special(name) >= 0)
-			fatal(_("make_builtin: can't use gawk built-in `%s' as function name"), name);
+			fatal(_("make_builtin: cannot use gawk built-in `%s' as function name"), name);
 
 		size_t len = strlen(name_space) + 2 + strlen(name) + 1;
 		char *buf;
-		emalloc(buf, char *, len, "make_builtin");
+		emalloc(buf, char *, len);
 		sprintf(buf, "%s::%s", name_space, name);
 		install_name = buf;
 
@@ -123,11 +174,16 @@ make_builtin(const char *name_space, const awk_ext_func_t *funcinfo)
 		// found it, but it shouldn't be there if we want to install this function
 		if (f->type == Node_func) {
 			/* user-defined function */
-			fatal(_("make_builtin: can't redefine function `%s'"), name);
+			fatal(_("make_builtin: cannot redefine function `%s'"), name);
 		} else if (f->type == Node_ext_func) {
 			/* multiple extension() calls etc. */
-			if (do_lint)
+			if (do_lint && ! using_persistent_malloc)
 				lintwarn(_("make_builtin: function `%s' already defined"), name);
+			free((void *) install_name);
+			if (using_persistent_malloc) {
+				f->code_ptr->extfunc = funcinfo->function;
+				return awk_true;
+			}
 			return awk_false;
 		} else
 			/* variable name etc. */
@@ -199,10 +255,15 @@ get_actual_argument(NODE *t, int i, bool want_array)
 	pc = TOP()->code_ptr;	/* Op_ext_builtin instruction */
 	fname = (pc + 1)->func_name;
 
-	if (t->type == Node_var_new) {
+	if (t->type == Node_var_new || t->type == Node_elem_new) {
 		if (want_array)
 			return force_array(t, false);
 		else {
+			if (t->type == Node_elem_new) {
+				elem_new_reset(t);
+				if (t->valref > 1)	// ADR: 2/2025: Can this happen?
+					unref(t);
+			}
 			t->type = Node_var;
 			t->var_value = dupnode(Nnull_string);
 			return t->var_value;
@@ -223,11 +284,17 @@ get_actual_argument(NODE *t, int i, bool want_array)
 }
 
 #else
+/* init_extension --- dummy version if extensions not available */
+
+void
+init_extension_list(void)
+{
+}
 
 /* load_ext --- dummy version if extensions not available */
 
 void
-load_ext(const char *lib_name)
+load_ext(const char *name, const char *lib_name)
 {
 	fatal(_("dynamic loading of libraries is not supported"));
 }
