@@ -35,6 +35,10 @@
 #include <spawn.h>	// for macos
 #endif
 
+#ifdef HAVE__NSGETEXECUTABLEPATH
+#include <mach-o/dyld.h>
+#endif
+
 const char quote = '\'';
 const char *defpath = DEFPATH;
 const char *deflibpath = DEFLIBPATH;
@@ -320,28 +324,21 @@ os_disable_aslr(const char *persist_file, char **argv)
 		const char *cp = getenv("GAWK_PMA_REINCARNATION");
 
 		if (cp == NULL) {
-			char fullpath[BUFSIZ];
-			int n;
+			static const char exe[] = "/proc/self/exe";
 
-			if ((n = readlink("/proc/self/exe", fullpath, sizeof(fullpath)-1)) < 0) {
-				fprintf(stderr, _("warning: /proc/self/exe: readlink: %s\n"),
-							strerror(errno));
-				return;
-			}
-			fullpath[n] = '\0';
-			putenv("GAWK_PMA_REINCARNATION=true");
+			(void) setenv("GAWK_PMA_REINCARNATION", "true", true);
 			if (personality(PER_LINUX | ADDR_NO_RANDOMIZE) < 0) {
 				fprintf(stderr, _("warning: personality: %s\n"),
 							strerror(errno));
 				fflush(stderr);
 				// do the exec anyway...
 			}
-			execv(fullpath, argv);
+			execv(exe, argv);
 		} else
 			(void) unsetenv("GAWK_PMA_REINCARNATION");
 	}
 #endif
-#ifdef HAVE__NSGETEXECUTABLEPATH
+#if defined(HAVE__NSGETEXECUTABLEPATH) && defined(HAVE_POSIX_SPAWNP)
 	// This code is for macos
 	if (persist_file != NULL) {
 		const char *cp = getenv("GAWK_PMA_REINCARNATION");
@@ -353,7 +350,7 @@ os_disable_aslr(const char *persist_file, char **argv)
 			int status;
 			pid_t pid;
 			extern char **environ;
-			size_t size = BUFSIZ;
+			uint32_t size = BUFSIZ;
 
 			memset(fullpath, 0, BUFSIZ);
 			n = _NSGetExecutablePath(fullpath, &size);
@@ -366,13 +363,13 @@ os_disable_aslr(const char *persist_file, char **argv)
 			if (status == 0) {
 				if (waitpid(pid, &status,  WUNTRACED) != -1) {
 					if (WIFEXITED(status))
-						exit WEXITSTATUS(status);	// use original exit code
+						exit(WEXITSTATUS(status));	// use original exit code
 				} else {
 					fprintf(stderr, _("waitpid: got exit status %#o\n"), status);
 					exit(EXIT_FATAL);
 				}
 			} else {
-				fprintf(stderr, _("fatal: posix_spawn: %s\n"), strerror(errno));
+				fprintf(stderr, _("fatal: posix_spawnp: %s\n"), strerror(errno));
 				exit(EXIT_FATAL);
 			}
 		} else
